<?php

namespace Drupal\frog\Controller;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Controller\ControllerBase;
use Drupal\frog\Entity\FrogDomain;
use Drupal\frog\Entity\FrogLink;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\frog\FrogLinkDownloaderInterface;

/**
 * Class FrogCrawlerController.
 */
class FrogCrawlerController extends ControllerBase {

  /**
   * Drupal\frog\LinkDownloaderInterface definition.
   *
   * @var \Drupal\frog\FrogLinkDownloaderInterface
   */
  protected $drupalFrogLinkDownloader;

  /**
   * FrogCrawlerController constructor.
   *
   * @param FrogLinkDownloaderInterface $frog_link_downloader
   */
  public function __construct(FrogLinkDownloaderInterface $frog_link_downloader) {
    $this->drupalFrogLinkDownloader = $frog_link_downloader;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('frog.link_downloader')
    );
  }

  /**
   * Scraping.
   *
   * @param \Drupal\frog\Entity\FrogDomain $frog_domain
   *   The domain.
   *
   * @return string
   *   Return Hello string.
   */
  public function scraping(FrogDomain $frog_domain) {
    $batch = [
      'title' => t('Scraping links...'),
      'operations' => [
        [
          [__CLASS__, 'checkDomain'], [$frog_domain]
        ],
      ],
      'finished' => [__CLASS__, 'scrapingFinishedCallback'],
      'init_message' => t('Preparing to scrap links...'),
    ];

    batch_set($batch);

    return batch_process('/admin/structure/domain');
  }

  /**
   * Check domain.
   *
   * @param \Drupal\frog\Entity\FrogDomain $domain
   *   The domain.
   * @param array $context
   *   Batch context.
   */
  public static function checkDomain(FrogDomain $domain, &$context) {
    $links = FrogCrawlerController::getLinks($domain);

    if (!isset($context['sandbox']['links'])) {
      $context['sandbox']['links'] = [];
      $context['results'] = [];
      $context['sandbox']['progress'] = 0;
      $context['sandbox']['current'] = 0;
    }

    // Without any links, the domain became the first link used to start scraping.
    if (!count($links)) {
      // Save the link with all the info updated.
      FrogCrawlerController::check($domain->getName(), $domain, '');
      $context['sandbox']['links'][$domain->getName()] = TRUE;
      $context['sandbox']['progress']++;
      $context['sandbox']['current']++;
      $context['message'] = 'Checking link ' . $domain->getName();
      FrogCrawlerController::setProgressMessage($context, $domain->getName());
    }
    else {
      $MAX_CHECK_BATCH = 2;
      $current = 1;
      foreach ($links as $link) {
        $link = FrogLink::load($link);
        if (isset($context['sandbox']['links'][$link->getURL()]) && $context['sandbox']['links'][$link->getURL()]) {
          continue;
        }
        //sleep(3);
        FrogCrawlerController::check($link->getUrl(), $domain, $domain->getName());
        $context['sandbox']['links'][$link->getUrl()] = TRUE;
        $context['sandbox']['progress']++;
        $context['sandbox']['current']++;
        FrogCrawlerController::setProgressMessage($context, $link->getURL());
        if ($current == $MAX_CHECK_BATCH) {
          break;
        }

        $current++;
      }
    }

    // Get all the link in the database to check if there are links that need to be
    // downloaded.
    // Consider that all the link downloaded has been passed as $link parameter
    // to the check() method and are saved into the $context['links'][$link] as TRUE.
    $links = FrogCrawlerController::getLinks($domain);
    foreach ($links as $link) {
      $link = FrogLink::load($link);
      if (!isset($context['sandbox']['links'][$link->getUrl()])) {
        $context['sandbox']['links'][$link->getUrl()] = FALSE;
      }
    }

    // Check if all the links has been checked.
    // Consider that all the link downloaded has been passed as $link parameter
    // to the check() method and are saved into the $context['links'][$link] as TRUE.
    $context['finished'] = TRUE;
    foreach ($context['sandbox']['links'] as $link) {
      if (!$link) {
        $context['finished'] = FALSE;
      }
    }

    // Update the amount pof links retrieved by the crawler.
    $context['results'] = $context['sandbox']['links'];


  }

  public function scrapingFinishedCallback($success, $results, $operations) {
    // The 'success' parameter means no fatal PHP errors were detected. All
    // other error management should be handled using 'results'.
    if ($success) {
      $message = \Drupal::translation()->formatPlural(
        count($results),
        'One post processed.', '@count posts processed.'
      );
    }
    else {
      $message = t('Finished with an error.');
    }
    drupal_set_message($message);
  }

  /**
   * Set progress message.
   *
   * @param mixed $context
   *   Context.
   * @param string $link_url
   *   Link url.
   */
  public static function setProgressMessage(&$context, $link_url) {
    $total_count = count($context['results']);
    $total = $total_count ? $total_count : 1;
    $percentage = round($context['sandbox']['current'] / $total * 100, 2);
    $context['message'] = t('Checking @current of @total (@percentage%). Link: @link', [
      '@current' => $context['sandbox']['current'],
      '@total' => $total,
      '@link' => $link_url,
      '@percentage' => $percentage,
    ]);
  }

  /**
   * @param string $link
   *   The link.
   * @param object $domain
   *   The domain.
   * @param string $parent_url
   *   The parent url.
   *
   * @return mixed
   */
  public static function check($link, $domain, $parent_url) {
    $link = FrogCrawlerController::prepareUrl($link, $domain);

    if (!UrlHelper::isValid($link, UrlHelper::isExternal($link))) {
      return FALSE;
    }

    if (!filter_var($link, FILTER_VALIDATE_URL)) {
      return FALSE;
    }

    // Save the link with all the info updated.
    \Drupal::service('frog.link_downloader')->save($link, $domain, $parent_url);

    if (strpos($link, $domain->getUrl()) !== false) {
      // Check if the link downloaded contain other links.
      $links = \Drupal::service('frog.link_downloader')->extractLinks();
      FrogCrawlerController::saveChildrenLinks($links, $domain, $link);
      return $links;
    }
  }

  /**
   * @return array|int
   */
  public function getDomains() {
    $query = \Drupal::entityQuery('frog_domain');
    // TODO: filter by older than x days or without an md5 version of the page.
    $domains = $query->execute();
    return $domains;
  }

  /**
   *
   */
  public static function getLinks($domain) {
    // Get all Link entities.
    $query = \Drupal::entityQuery('frog_link');
    // TODO: filter by older than x days or without an md5 version of the page.
    //$query->condition('domain', $domain);
    $links = $query->execute();
    return $links;
  }

  /**
   * @param $url
   * @param $domain
   *
   * @return string
   */
  public static function prepareUrl($url, $domain) {
    // Create a new link with the url and the status code.
    if (!UrlHelper::isExternal($url)) {
      if (strpos($url, '/') === FALSE) {
        // then the url is relative and need to be completed with the domain.
        $url = '/' . $url;
      }

      $url = rtrim($domain->getUrl(), '/') . $url;
    }

    return $url;
  }

  /**
   * Save children links.
   *
   * @param $links
   *   The children links.
   * @param object $domain
   *   The domain.
   * @param $parent_url
   *   The parent url, if exists.
   */
  public static function saveChildrenLinks($links, $domain, $parent_url) {
    foreach ($links as $url) {
      $url = FrogCrawlerController::prepareUrl($url, $domain);

      if (UrlHelper::isValid($url)) {
        \Drupal::service('frog.link_downloader')->save($url, $domain, $parent_url, FALSE);
      }
    }
  }
}
